from datetime import datetime

import pytz
from django.test import TestCase
from django.utils.timezone import now

from slick_reporting.generator import ReportGenerator
from slick_reporting.helpers import get_foreign_keys
from .models import OrderLine

from .report_generators import GeneratorWithAttrAsColumn, CrosstabOnClient

from .tests import BaseTestData
from .models import SimpleSales


class MatrixTests(BaseTestData, TestCase):
    def test_matrix_column_included(self):
        report = CrosstabOnClient(crosstab_ids=[self.client1.pk], crosstab_compute_reminder=False)
        columns = report.get_list_display_columns()
        self.assertEqual(len(columns), 3, columns)

        report = CrosstabOnClient(crosstab_ids=[self.client1.pk], crosstab_compute_reminder=True)
        columns = report.get_list_display_columns()
        self.assertEqual(len(columns), 4, columns)

    def test_get_crosstab_columns(self):
        report = CrosstabOnClient(crosstab_ids=[self.client1.pk])
        columns = report.get_list_display_columns()
        self.assertEqual(len(columns), 4)

        report = CrosstabOnClient(crosstab_ids=[self.client1.pk, self.client2.pk],
                                  crosstab_columns=['__total_quantity__', '__balance_quan__'])
        columns = report.get_list_display_columns()
        self.assertEqual(len(columns), 8, [x['name'] for x in columns])


class GeneratorReportStructureTest(TestCase):
    def test_time_series_columns_inclusion(self):
        x = ReportGenerator(OrderLine, date_field='order__date_placed', group_by='client', columns=['name'],
                            time_series_columns=['__total_quantity__'], time_series_pattern='monthly',
                            start_date=datetime(2020, 1, 1, tzinfo=pytz.timezone('utc')),
                            end_date=datetime(2020, 12, 31, tzinfo=pytz.timezone('utc')))
        # import pdb;
        # pdb.set_trace()
        self.assertEqual(len(x.get_list_display_columns()), 13)

    def test_time_series(self):
        pass

    def test_cross_tab(self):
        pass

    def test_time_series_and_cros_tab(self):
        pass

    def test_attr_as_column(self):
        report = GeneratorWithAttrAsColumn()
        columns_data = report.get_list_display_columns()
        self.assertEqual(len(columns_data), 3)

        self.assertEqual(columns_data[0]['verbose_name'], 'get_data_verbose_name')

    def test_improper_group_by(self):
        def load():
            ReportGenerator(OrderLine, group_by='no_field', date_field='order__date_placed')

        self.assertRaises(Exception, load)

    def _test_attr_called(self):
        report = GeneratorWithAttrAsColumn()
        data = report.get_report_data()
        # self.assertEqual(len(report.get_list_display_columns()), 3)

    def test_gather_dependencies_for_time_series(self):
        report = ReportGenerator(report_model=SimpleSales, group_by='client',
                                 columns=['slug', 'name'],
                                 time_series_pattern='monthly',
                                 date_field='doc_date',
                                 time_series_columns=['__debit__', '__credit__', '__balance__', '__total__']
                                 )

        self.assertTrue(report._report_fields_dependencies)


# test that columns are a straight forward list
class TestReportFields(TestCase):

    def test_get_full_dependency_list(self):
        from slick_reporting.fields import BalanceReportField
        deps = BalanceReportField.get_full_dependency_list()
        self.assertEqual(len(deps), 1)


class TestHelpers(TestCase):
    def test_get_model_for_keys(self):
        keys = get_foreign_keys(OrderLine)
        self.assertEqual(len(keys), 3)



from collections import OrderedDict

from django.db import models


def get_calculation_annotation(calculation_field, calculation_method):
    '''
    Returns the default django annotation
    @param calculation_field: the field to calculate ex 'value'
    @param calculation_method: the aggregation method ex: Sum
    @return: the annotation ex value__sum
    '''

    return '__'.join([calculation_field.lower(), calculation_method.name.lower()])


def get_foreign_keys(model):
    fields = model._meta.get_fields()
    fkeys = OrderedDict()
    for f in fields:
        if f.is_relation and type(f) is not models.OneToOneRel \
                and type(f) is not models.ManyToOneRel and type(f) is not models.ManyToManyRel:
            fkeys[f.attname] = f

    return fkeys


def get_field_from_query_text(path, model):
    relations = path.split('__')
    _rel = model
    field = None
    for i, m in enumerate(relations):
        field = _rel._meta.get_field(m)
        if i == len(relations) - 1:
            return field
        _rel = field.related_model
    return field



from django import apps



class ReportAppConfig(apps.AppConfig):
    label = 'Slick Reporting'
    name = 'slick_reporting'

    def ready(self):
        super().ready()

        from . import fields



import datetime
from unittest import skip
from urllib.parse import urljoin

from django.contrib.auth import get_user_model
from django.test import SimpleTestCase, TestCase, override_settings
from django.urls import reverse
from django.utils.timezone import now

from slick_reporting.generator import ReportGenerator
from tests.report_generators import ClientTotalBalance
from .models import Client, Product, SimpleSales, OrderLine

User = get_user_model()
SUPER_LOGIN = dict(username='superlogin', password='password')
year = now().year

from . import report_generators


class ReportRegistryTest(SimpleTestCase):
    pass


class BaseTestData:
    databases = '__all__'

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        User.objects.create_superuser('super', None, 'secret')

        user = User.objects.create(is_superuser=True, is_staff=True, **SUPER_LOGIN)
        limited_user = User.objects.create_user(is_superuser=False, is_staff=True, username='limited',
                                                password='password')
        cls.user = user
        cls.limited_user = limited_user
        cls.client1 = Client.objects.create(name='Client 1')
        cls.client2 = Client.objects.create(name='Client 2')
        cls.client3 = Client.objects.create(name='Client 3')
        cls.clientIdle = Client.objects.create(name='Client Idle')

        cls.product1 = Product.objects.create(name='Product 1')
        cls.product2 = Product.objects.create(name='Product 2')
        cls.product3 = Product.objects.create(name='Product 3')

        SimpleSales.objects.create(
            doc_date=datetime.datetime(year, 1, 2), client=cls.client1,
            product=cls.product1, quantity=10, price=10)
        SimpleSales.objects.create(
            doc_date=datetime.datetime(year, 2, 2), client=cls.client1,
            product=cls.product1, quantity=10, price=10)

        SimpleSales.objects.create(
            doc_date=datetime.datetime(year, 3, 2), client=cls.client1,
            product=cls.product1, quantity=10, price=10)

        # client 2
        SimpleSales.objects.create(
            doc_date=datetime.datetime(year, 1, 2), client=cls.client2,
            product=cls.product1, quantity=20, price=10)
        SimpleSales.objects.create(
            doc_date=datetime.datetime(year, 2, 2), client=cls.client2,
            product=cls.product1, quantity=20, price=10)

        SimpleSales.objects.create(
            doc_date=datetime.datetime(year, 3, 2), client=cls.client2,
            product=cls.product1, quantity=20, price=10)

        # client 3
        SimpleSales.objects.create(
            doc_date=datetime.datetime(year, 1, 2), client=cls.client3,
            product=cls.product1, quantity=30, price=10)
        SimpleSales.objects.create(
            doc_date=datetime.datetime(year, 2, 2), client=cls.client3,
            product=cls.product1, quantity=30, price=10)

        SimpleSales.objects.create(
            doc_date=datetime.datetime(year, 3, 2), client=cls.client3,
            product=cls.product1, quantity=30, price=10)


# @override_settings(ROOT_URLCONF='reporting_tests.urls', RA_CACHE_REPORTS=False, USE_TZ=False)
class ReportTest(BaseTestData, TestCase):

    def test_client_balance(self):
        report = report_generators.ClientTotalBalance()
        data = report.get_report_data()

        self.assertEqual(data[0].get('__balance__'), 300, data[0])

    def test_product_total_sales(self):
        report = report_generators.ProductTotalSales()
        data = report.get_report_data()
        self.assertEqual(data[0]['__balance__'], 1800)

    def test_client_client_sales_monthly(self):
        report = report_generators.ClientSalesMonthlySeries()

        data = report.get_report_data()

        self.assertEqual(data[0].get('__balance__TS%s0301' % year), 200, data[0])
        self.assertEqual(data[0]['__balance__TS%s0201' % year], 100)

        self.assertEqual(data[0]['__total__TS%s0401' % year], 100)
        self.assertEqual(data[0]['__total__TS%s0301' % year], 100)
        self.assertEqual(data[0]['__total__TS%s0201' % year], 100)

        # todo add __fb__ to time series and check the balance

    def test_client_statement_detail(self):
        """
        Test the detail statement
        This is do pass by making a document slug clickable (<a> elem)
        and it also passes by the slug search of the model admin
        :return:
        """
        report = report_generators.ClientDetailedStatement()
        data = report.get_report_data()
        self.assertEqual(len(data), 9)

    def test_productclientsalesmatrix(self):
        report = report_generators.ProductClientSalesMatrix(crosstab_ids=[self.client1.pk, self.client2.pk])
        data = report.get_report_data()
        self.assertEqual(data[0]['__total__CT%s' % self.client1.pk], 300)
        self.assertEqual(data[0]['__total__CT%s' % self.client2.pk], 600)
        self.assertEqual(data[0]['__total__CT----'], 900)

    def _test_default_order_by(self):
        self.client.login(username='super', password='secret')
        response = self.client.get(reverse('ra_admin:report', args=('client', 'clienttotalbalancesordered')),
                                   HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)
        data = response.json()
        previous_balance = 0
        self.assertTrue(len(data['data']) > 1)
        for i, line in enumerate(data['data']):
            if i == 0:
                previous_balance = line['__balance__']
            else:
                self.assertTrue(line['__balance__'] > previous_balance)

    def _test_default_order_by_reversed(self):
        self.client.login(username='super', password='secret')
        response = self.client.get(reverse('ra_admin:report', args=('client', 'ClientTotalBalancesOrderedDESC')),
                                   HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)
        data = response.json()
        previous_balance = 0
        self.assertTrue(len(data['data']) > 1)
        for i, line in enumerate(data['data']):
            if i == 0:
                previous_balance = line['__balance__']
            else:
                self.assertTrue(line['__balance__'] < previous_balance)

    def test_show_empty_records(self):
        report = report_generators.ClientTotalBalance()
        data = report.get_report_data()
        with_show_empty_len = len(data)
        wo_show_empty = report_generators.ClientTotalBalance(show_empty_records=False)
        self.assertNotEqual(with_show_empty_len, wo_show_empty)
        # self.assertEqual(data[0].get('__balance__'), 300, data[0])

    def test_filters(self):
        report = ClientTotalBalance(kwargs_filters={'client': self.client1.pk}, show_empty_records=True)
        data = report.get_report_data()
        self.assertEqual(len(data), 1, data)

        report = ClientTotalBalance(kwargs_filters={'client': self.client1.pk}, show_empty_records=False)
        data = report.get_report_data()
        self.assertEqual(len(data), 1, data)

    def test_filter_as_int_n_list(self):
        report = ClientTotalBalance(kwargs_filters={'client': self.client1.pk}, show_empty_records=True)
        data = report.get_report_data()
        self.assertEqual(len(data), 1, data)

        report = ClientTotalBalance(kwargs_filters={'client_id__in': [self.client1.pk]}, show_empty_records=True)
        data = report.get_report_data()
        self.assertEqual(len(data), 1, data)


class TestView(BaseTestData, TestCase):
    def test_view(self):
        reponse = self.client.get(reverse('report1'))
        self.assertEqual(reponse.status_code, 200)
        # import pdb; pdb.set_trace()
        view_report_data = reponse.context['report_data']['data']
        report_generator = ReportGenerator(report_model=OrderLine,
                                           date_field='date_placed',  # or 'order__date_placed',
                                           group_by='product',
                                           columns=['name', 'sku'],
                                           time_series_pattern='monthly',
                                           time_series_columns=['__total_quantity__'],
                                           )
        self.assertEqual(view_report_data, report_generator.get_report_data())



import pytz
from django.conf import settings
from django.utils.functional import lazy

import datetime


def get_first_of_this_year():
    d = datetime.datetime.today()
    return datetime.datetime(d.year, 1, 1, 0, 0, 0, 0, pytz.timezone(settings.TIME_ZONE))


def get_end_of_this_year():
    d = datetime.datetime.today()
    return datetime.datetime(d.year + 1, 1, 1, 0, 0, 0, 0, pytz.timezone(settings.TIME_ZONE))


SLICK_REPORTING_DEFAULT_START_DATE = getattr(settings, '', lazy(get_first_of_this_year, datetime.datetime)())
SLICK_REPORTING_DEFAULT_END_DATE = getattr(settings, '', lazy(get_end_of_this_year, datetime.datetime)())



import os

SECRET_KEY = 'fake-key'

PROJECT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

BASE_DIR = os.path.dirname(PROJECT_DIR)

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    },
}

PASSWORD_HASHERS = [
    'django.contrib.auth.hashers.MD5PasswordHasher',
]

INSTALLED_APPS = [
    # 'django.contrib.admin',
    'django.contrib.contenttypes',
    'django.contrib.auth',
    'django.contrib.sites',
    'django.contrib.sessions',
    'django.contrib.messages',


    # 'django.contrib.admin.apps.SimpleAdminConfig',
    'django.contrib.staticfiles',
    'slick_reporting',
    'crispy_forms',
    "tests",
]

ROOT_URLCONF = 'tests.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'django.template.context_processors.static',

            ],
        },
    },
]
STATIC_URL = '/static/'



from collections import OrderedDict

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Column
from django import forms
from django.utils.translation import ugettext_lazy as _

from . import app_settings
from .helpers import get_foreign_keys


class RaReportBaseForm(object):
    '''
    Holds basic function
    '''

    def get_filters(self):
        """
        Get the foreign key filters for report queryset,
        :return: a dicttionary of filters to be used with QuerySet.filter(**returned_value)
        """
        # todo: implement cross tab support
        _values = {}
        if self.is_valid():
            for key, field in self.foreign_keys.items():
                if key in self.cleaned_data:
                    val = self.cleaned_data[key]
                    if val:
                        val = [x for x in val.values_list('pk', flat=True)]
                        _values['%s__in' % key] = val
            return None, _values

    def get_crispy_helper(self, foreign_keys_map=None, crosstab_model=None, **kwargs):

        helper = FormHelper()
        helper.form_class = 'form-horizontal'
        helper.label_class = 'col-sm-2 col-md-2 col-lg-2'
        helper.field_class = 'col-sm-10 col-md-10 col-lg-10'
        helper.form_tag = False
        helper.disable_csrf = True

        foreign_keys_map = foreign_keys_map or self.foreign_keys
        from crispy_forms.layout import Layout, Div, Row, Field
        # from crispy_forms.bootstrap import

        helper.layout = Layout(
            # PanelContainer(
            # #     Div(
            #         _('filters'),
            Row(
                # Div(StackedField('doc_date'), css_class='col-sm-3'),
                # InlineField('start_date', css_class='form-control dateinput'), css_class='col-sm-6'),
                # Div(InlineField('end_date', css_class='form-control dateinput'), css_class='col-sm-6'),
                Column(
                    Field('start_date'), css_class='col-sm-6'),
                Column(
                    Field('end_date'), css_class='col-sm-6'),

                css_class='raReportDateRange'),
            Div(css_class="mt-20", style='margin-top:20px')
        )

        # if crosstab_model:
        #     entry_point.append(Row(
        #         Div('matrix_entities', css_class='col-sm-9'),
        #         Div('matrix_show_other', css_class='col-sm-3')
        #         , css_class='matrixField')
        #     )

        for k in foreign_keys_map:
            if k[:-3] != crosstab_model:
                helper.layout.fields[1].append(Field(k))

        return helper


def _default_foreign_key_widget(f_field):
    # import pdb; pdb.set_trace()
    return {'form_class': forms.ModelMultipleChoiceField,
            'required': False,
            # 'widget': RaAutocompleteSelectMultiple(f_field.remote_field, ra_admin_site,
            #                                        attrs={'class': 'select2bs4'})
            }


def report_form_factory(model, fkeys_filter_func=None, foreign_key_widget_func=None, **kwargs):
    foreign_key_widget_func = foreign_key_widget_func or _default_foreign_key_widget
    fkeys_filter_func = fkeys_filter_func or (lambda x: x)

    # gather foreign keys
    fkeys_map = get_foreign_keys(model)
    fkeys_map = fkeys_filter_func(fkeys_map)

    fkeys_list = []
    fields = OrderedDict()

    fields['start_date'] = forms.SplitDateTimeField(required=False, label=_('From date'),
                                                    initial=app_settings.SLICK_REPORTING_DEFAULT_START_DATE,
                                                    # widget=RaBootstrapDateTime(),
                                                    input_date_formats=['%Y-%m-%d', '%Y-%m-%d'],
                                                    input_time_formats=['%H:%M', '%H:%M:%S'],
                                                    )

    fields['end_date'] = forms.SplitDateTimeField(required=False, label=_('To  date'),
                                                  initial=app_settings.SLICK_REPORTING_DEFAULT_END_DATE,
                                                  # widget=RaBootstrapDateTime(),
                                                  )

    for name, f_field in fkeys_map.items():
        fkeys_list.append(name)

        fields[name] = f_field.formfield(
            **foreign_key_widget_func(f_field))

    new_form = type('ReportForm', (RaReportBaseForm, forms.BaseForm,),
                    {"base_fields": fields,
                     '_fkeys': fkeys_list,
                     'foreign_keys': fkeys_map,
                     })
    return new_form



from __future__ import unicode_literals

from django import forms


class OrderByForm(forms.Form):
    order_by = forms.CharField(required=False)

    def get_order_by(self, default_field=None):
        """
        Get the order by specified by teh form or the default field if provided
        :param default_field:
        :return: tuple of field and direction
        """
        if self.is_valid():
            order_field = self.cleaned_data['order_by']
            order_field = order_field or default_field
            if order_field:
                return self.parse_order_by_field(order_field)
        return None, None

    def parse_order_by_field(self, order_field):
        """
        Specify the field and direction
        :param order_field: the field to order by
        :return: tuple of field and direction
        """
        if order_field:
            asc = True
            if order_field[0:1] == '-':
                order_field = order_field[1:]
                asc = False
            return order_field, not asc
        return None, None



from __future__ import unicode_literals

from django.contrib.admin.sites import AlreadyRegistered, NotRegistered


class ReportFieldRegistry(object):
    def __init__(self):
        super(ReportFieldRegistry, self).__init__()
        self._registry = {}  # holds

    def register(self, report_field):
        if report_field.name in self._registry:
            raise AlreadyRegistered('This field is already registered')
        self._registry[report_field.name] = report_field
        return report_field

    def unregister(self, report_field):
        if report_field.name not in self._registry:
            raise NotRegistered(report_field)
        del self._registry[report_field.name]

    def get_field_by_name(self, name):
        if name in self._registry:
            return self._registry[name]
        else:
            raise NotRegistered(name)

    def get_all_report_fields_names(self):
        return list(self._registry.keys())


field_registry = ReportFieldRegistry()





from __future__ import unicode_literals

import datetime
import logging
from collections import Iterable
from inspect import isclass

from django.core.exceptions import ImproperlyConfigured, FieldDoesNotExist
from django.db.models import Q

from .helpers import get_field_from_query_text
from .registry import field_registry

logger = logging.getLogger(__name__)


class ReportGenerator(object):
    """
    The main class responsible generating the report and managing the flow
    """

    field_registry_class = field_registry
    """You can have a custom computation field locator! It only needs a `get_field_by_name(string)` 
    and returns a ReportField`"""

    report_model = None
    """The main model where data is """

    """
    Class to generate a Json Object containing report data.
    """
    date_field = None
    """Main date field to use whenever date filter is needed"""

    print_flag = None
    list_display_links = []

    group_by = None
    """The field to use for grouping, if not set then the report is expected to be a sub version of the report model"""

    columns = None
    """A list of column names.
    Columns names can be 
    
    1. A Computation Field
    
    2. If group_by is set, then any field on teh group_by model
    
    3. If group_by is not set, then any field name on the report_model / queryset
     
    4. A callable on the generator
      
    5. Special __time_series__, and __crosstab__ 
       Those can be use to control the position of the time series inside the columns, defaults it's appended at the end
       
       Example:
       columns = ['product_id', '__time_series__', 'col_b']
       Same is true with __crosstab__ 
     """

    time_series_pattern = ''
    """
    If set the Report will compute a time series.
    
    Possible options are: daily, weekly, semimonthly, monthly, quarterly, semiannually, annually and custom.
    
    if `custom` is set, you'd need to override  `get_custom_time_series_dates`
    """
    time_series_columns = None
    """
    a list of Calculation Field names which will be included in the series calculation.
     Example: ['__total__', '__total_quantity__'] with compute those 2 fields for all the series
     
    """

    crosstab_model = None
    """
    If set, a cross tab over this model selected ids (via `crosstab_ids`)  
    """
    crosstab_columns = None
    """The computation fields which will be computed for each crosstab-ed ids """

    crosstab_ids = None
    """A list is the ids to create a crosstab report on"""

    crosstab_compute_reminder = True
    """Include an an extra crosstab_columns for the outer group ( ie: all expects those `crosstab_ids`) """

    show_empty_records = True
    """
    If group_by is set, this option control if the report result will include all objects regardless of appearing in the report_model/qs.
    If set False, only those objects which are found in the report_model/qs
    Example: Say you group by client
    show_empty_records = True will get the computation fields for all clients in the Client model (including those who 
    didnt make a transaction.
    
    show_empty_records = False will get the computation fields for all clients in the Client model (including those who 
    didnt make a transaction.
     
    """

    limit_records = None
    """Serves are a main limit to  the returned data of teh report_model.
    Can be beneficial if the results may be huge.
    """
    swap_sign = False

    def __init__(self, report_model=None, main_queryset=None, start_date=None, end_date=None, date_field=None,
                 q_filters=None, kwargs_filters=None,
                 group_by=None, columns=None,
                 time_series_pattern=None, time_series_columns=None,
                 crosstab_model=None, crosstab_columns=None, crosstab_ids=None, crosstab_compute_reminder=None,
                 swap_sign=False, show_empty_records=None,
                 print_flag=False,
                 doc_type_plus_list=None, doc_type_minus_list=None, limit_records=False, ):
        """

        :param report_model: Main model containing the data
        :param main_queryset: Default to report_model.objects
        :param start_date:
        :param end_date:
        :param date_field:
        :param q_filters:
        :param kwargs_filters:
        :param group_by:
        :param columns:
        :param time_series_pattern:
        :param time_series_columns:
        :param crosstab_model:
        :param crosstab_columns:
        :param crosstab_ids:
        :param crosstab_compute_reminder:
        :param swap_sign:
        :param show_empty_records:
        :param base_model:
        :param print_flag:
        :param doc_type_plus_list:
        :param doc_type_minus_list:
        :param limit_records:
        """
        from .app_settings import SLICK_REPORTING_DEFAULT_START_DATE, SLICK_REPORTING_DEFAULT_END_DATE

        super(ReportGenerator, self).__init__()

        self.report_model = self.report_model or report_model
        if not self.report_model:
            raise ImproperlyConfigured('report_model must be set on a class level or via init')

        self.start_date = start_date or datetime.datetime.combine(SLICK_REPORTING_DEFAULT_START_DATE.date(),
                                                                  SLICK_REPORTING_DEFAULT_START_DATE.time())

        self.end_date = end_date or datetime.datetime.combine(SLICK_REPORTING_DEFAULT_END_DATE.date(),
                                                              SLICK_REPORTING_DEFAULT_END_DATE.time())
        self.date_field = self.date_field or date_field
        if not self.date_field:
            raise ImproperlyConfigured('date_field must be set on a class level or via init')

        self.q_filters = q_filters or []
        self.kwargs_filters = kwargs_filters or {}

        self.crosstab_model = self.crosstab_model or crosstab_model
        self.crosstab_columns = crosstab_columns or self.crosstab_columns or []
        self.crosstab_ids = self.crosstab_ids or crosstab_ids or []
        self.crosstab_compute_reminder = self.crosstab_compute_reminder if crosstab_compute_reminder is None else crosstab_compute_reminder

        main_queryset = main_queryset or self.report_model.objects

        self.columns = self.columns or columns or []
        self.group_by = self.group_by or group_by

        self.time_series_pattern = self.time_series_pattern or time_series_pattern
        self.time_series_columns = self.time_series_columns or time_series_columns

        self._prepared_results = {}
        self.report_fields_classes = {}

        self._report_fields_dependencies = {'time_series': {}, 'crosstab': {}, 'normal': {}}
        self.existing_dependencies = {'series': [], 'matrix': [], 'normal': []}

        self.print_flag = print_flag or self.print_flag

        # todo validate columns is not empty (if no time series / cross tab)

        if self.group_by:
            try:
                self.group_by_field = [x for x in self.report_model._meta.fields if x.name == self.group_by][0]
            except IndexError:
                raise ImproperlyConfigured(
                    f'Can not find group_by field:{self.group_by} in report_model {self.report_model} ')

            self.focus_field_as_key = self.group_by
            self.group_by_field_attname = self.group_by_field.attname
        else:
            self.focus_field_as_key = None
            self.group_by_field_attname = None

        # doc_types = form.get_doc_type_plus_minus_lists()
        doc_types = [], []
        self.doc_type_plus_list = list(doc_type_plus_list) if doc_type_plus_list else doc_types[0]
        self.doc_type_minus_list = list(doc_type_minus_list) if doc_type_minus_list else doc_types[1]

        self.swap_sign = self.swap_sign or swap_sign
        self.limit_records = self.limit_records or limit_records

        # passed to the report fields
        # self.date_field = date_field or self.date_field

        # in case of a group by, do we show a grouped by model data regardless of their appearance in the results
        # a client who didnt make a transaction during the date period.
        self.show_empty_records = False # show_empty_records if show_empty_records else self.show_empty_records
        # Looks like this options is harder then what i thought as it interfere with the usual filtering of the report

        # Preparing actions
        self._parse()
        if self.group_by:

            if self.show_empty_records:
                pass
                # group_by_filter = self.kwargs_filters.get(self.group_by, '')
                # qs = self.group_by_field.related_model.objects
                # if group_by_filter:
                #     import pdb; pdb.set_trace()
                #     lookup = 'pk__in' if isinstance(group_by_filter, Iterable) else 'pk'
                #     qs = qs.filter(**{lookup: group_by_filter})
                # self.main_queryset = qs.values()

            else:
                self.main_queryset = self._apply_queryset_options(main_queryset)
                ids = self.main_queryset.values_list(self.group_by_field.attname).distinct()
                self.main_queryset = self.group_by_field.related_model.objects.filter(pk__in=ids).values()
        else:
            self.main_queryset = self._apply_queryset_options(main_queryset, self.get_database_columns())

        self._prepare_report_dependencies()

    def _apply_queryset_options(self, query, fields=None):
        """
        Apply the filters to the main queryset which will computed results be mapped to
        :param query:
        :param fields:
        :return:
        """

        filters = {
            f'{self.date_field}__gt': self.start_date,
            f'{self.date_field}__lte': self.end_date,
        }
        filters.update(self.kwargs_filters)

        if filters:
            query = query.filter(**filters)
        if fields:
            return query.values(*fields)
        return query.values()

    def _construct_crosstab_filter(self, col_data):
        """
        In charge of adding the needed crosstab filter, specific to the case of is_reminder or not
        :param col_data:
        :return:
        """
        if col_data['is_reminder']:
            filters = [~Q(**{f"{col_data['model']}_id__in": self.crosstab_ids})]
        else:
            filters = [Q(**{f"{col_data['model']}_id": col_data['id']})]
        return filters

    def _prepare_report_dependencies(self):
        from .fields import BaseReportField
        all_columns = (
            ('normal', self._parsed_columns),
            ('time_series', self._time_series_parsed_columns),
            ('crosstab', self._crosstab_parsed_columns),
        )
        for window, window_cols in all_columns:
            for col_data in window_cols:
                klass = col_data['ref']
                dependencies_names = []
                # try:

                if isclass(klass) and issubclass(klass, BaseReportField):
                    dependencies_names = klass.get_full_dependency_list()

                    # check if any of this dependencies is on the report
                    fields_on_report = [x for x in window_cols if x['ref'] in dependencies_names]
                    for field in fields_on_report:
                        self._report_fields_dependencies[window][field['name']] = col_data['name']
            # import pdb; pdb.set_trace()
            for col_data in window_cols:
                klass = col_data['ref']
                # if getattr(klass, 'name', '') not in klasses_names:
                #     continue
                name = col_data['name']

                # if column has a dependency then skip it
                if not (isclass(klass) and issubclass(klass, BaseReportField)):
                    continue
                if self._report_fields_dependencies[window].get(name, False):
                    continue

                report_class = klass(self.doc_type_plus_list, self.doc_type_minus_list,
                                     group_by=self.group_by,
                                     report_model=self.report_model, date_field=self.date_field)

                q_filters = None
                date_filter = {
                    f'{self.date_field}__gt': col_data.get('start_date', self.start_date),
                    f'{self.date_field}__lte': col_data.get('end_date', self.end_date),
                }
                date_filter.update(self.kwargs_filters)
                if window == 'crosstab':
                    q_filters = self._construct_crosstab_filter(col_data)

                # print(f'preparing {report_class} for {window}')
                report_class.prepare(q_filters, date_filter)
                self.report_fields_classes[name] = report_class

    def _get_record_data(self, obj, columns):
        """
        the function is run for every obj in the main_queryset
        :param obj: current row
        :param: columns： The columns we iterate on
        :return: a dict object containing all needed data
        """

        # todo , if columns are empty for whatever reason this will throw an error
        display_link = self.list_display_links or columns[0]
        data = {}
        group_by_val = None
        if self.group_by:
            column_data = obj.get(self.group_by_field_attname, obj.get('id'))
            group_by_val = str(column_data)

        for window, window_cols in columns:
            for col_data in window_cols:

                name = col_data['name']

                if name.startswith('__') and self.group_by:
                    source = self._report_fields_dependencies[window].get(name, False)
                    if source:
                        computation_class = self.report_fields_classes[source]
                        value = computation_class.get_dependency_value(group_by_val,
                                                                       col_data['ref'].name)
                    else:
                        try:
                            computation_class = self.report_fields_classes[name]
                        except KeyError:
                            continue
                        value = computation_class.resolve(group_by_val)
                    if self.swap_sign: value = -value
                    data[name] = value

                else:
                    data[name] = obj.get(name, '')
                # if self.group_by and name in display_link:
                #     data[name] = make_linkable_field(self.group_by_field.related_model, group_by_val, data[name])
        return data

    def get_report_data(self):
        main_queryset = self.main_queryset[:self.limit_records] if self.limit_records else self.main_queryset

        all_columns = (
            ('normal', self._parsed_columns),
            ('time_series', self._time_series_parsed_columns),
            ('crosstab', self._crosstab_parsed_columns),
        )

        get_record_data = self._get_record_data
        data = [get_record_data(obj, all_columns) for obj in main_queryset]
        return data

    def _parse(self):

        if self.group_by:
            self.group_by_field = [x for x in self.report_model._meta.fields if x.name == self.group_by][0]
            self.group_by_model = self.group_by_field.related_model

        self.parsed_columns = []
        for col in self.columns:
            # import pdb; pdb.set_trace()
            attr = getattr(self, col, None)
            if attr:
                col_data = {'name': col,
                            'verbose_name': getattr(attr, 'verbose_name', col),
                            # 'type': 'method',
                            'ref': attr,
                            'type': 'text'
                            }
            elif col.startswith('__'):
                # a magic field
                if col in ['__time_series__', '__crosstab__']:
                    #     These are placeholder not real computation field
                    continue

                magic_field_class = field_registry.get_field_by_name(col)
                col_data = {'name': col,
                            'verbose_name': magic_field_class.verbose_name,
                            'source': 'magic_field',
                            'ref': magic_field_class,
                            'type': magic_field_class.type}
            else:
                # A database field
                model_to_use = self.group_by_model if self.group_by else self.report_model
                try:
                    if '__' in col:
                        # A traversing link order__client__email
                        field = get_field_from_query_text(col, model_to_use)
                    else:
                        field = model_to_use._meta.get_field(col)
                except FieldDoesNotExist:
                    raise FieldDoesNotExist(
                        f'Field "{col}" not found as an attribute to the generator class, nor as computation field, nor as a database column for the model "{model_to_use._meta.model_name}"')

                col_data = {'name': col,
                            'verbose_name': field.verbose_name,
                            'source': 'database',
                            'ref': field,
                            'type': field.get_internal_type()
                            }
            self.parsed_columns.append(col_data)

            self._parsed_columns = list(self.parsed_columns)
            self._time_series_parsed_columns = self.get_time_series_parsed_columns()
            self._crosstab_parsed_columns = self.get_crosstab_parsed_columns()

    def get_database_columns(self):
        return [col['name'] for col in self.parsed_columns if col['source'] == 'database']

    def get_method_columns(self):
        return [col['name'] for col in self.parsed_columns if col['type'] == 'method']

    def get_list_display_columns(self):
        columns = self.parsed_columns
        if self.time_series_pattern:
            time_series_columns = self.get_time_series_parsed_columns()
            try:
                index = self.columns.index('__time_series__')
                columns[index] = time_series_columns
            except ValueError:
                columns += time_series_columns

        if self.crosstab_model:
            crosstab_columns = self.get_crosstab_parsed_columns()

            try:
                index = self.columns.index('__crosstab__')
                columns[index] = crosstab_columns
            except ValueError:
                columns += crosstab_columns

        return columns

    def get_time_series_parsed_columns(self):
        """
        Return time series columns with all needed data attached
        :param plain: if True it returns '__total__' instead of '__total_TS011212'
        :return: List if columns
        """
        _values = []

        cols = self.time_series_columns or []
        series = self._get_time_series_dates()

        for dt in series:
            for col in cols:
                try:
                    magic_field_class = field_registry.get_field_by_name(col)
                except:
                    magic_field_class = None

                _values.append({
                    'name': col + 'TS' + dt[1].strftime('%Y%m%d'),
                    'original_name': col,
                    'verbose_name': self.get_time_series_field_verbose_name(magic_field_class, dt),
                    'ref': magic_field_class,
                    'start_date': dt[0],
                    'end_date': dt[1],
                })
        return _values

    def get_time_series_field_verbose_name(self, computation_class, date_period):
        """
        Sent the column data to construct a verbose name.
        Default implemenetation is column name + the end date %Y%m%d
        :param column_name: the computation field_name
        :param date_period: a tuple of (start_date, end_date)
        :return: a verbose string
        """
        return computation_class.get_time_series_field_verbose_name(date_period)

    def get_custom_time_series_dates(self):
        """
        Hook to get custom , maybe separated date periods
        :return: [ (date1,date2) , (date3,date4), .... ]
        """
        return []

    def _get_time_series_dates(self):
        from dateutil.relativedelta import relativedelta
        _values = []
        series = self.time_series_pattern
        if series:
            if series == 'daily':
                time_delta = datetime.timedelta(days=1)
            elif series == 'weekly':
                time_delta = relativedelta(weeks=1)
            elif series == 'semimonthly':
                time_delta = relativedelta(weeks=2)
            elif series == 'monthly':
                time_delta = relativedelta(months=1)
            elif series == 'quarterly':
                time_delta = relativedelta(months=3)
            elif series == 'semiannually':
                time_delta = relativedelta(months=6)
            elif series == 'annually':
                time_delta = relativedelta(year=1)
            elif series == 'custom':
                return self.get_custom_time_series_dates()
            else:
                raise NotImplementedError()

            done = False
            start_date = self.start_date

            # import pdb; pdb.set_trace()
            while not done:
                to_date = start_date + time_delta
                _values.append((start_date, to_date))
                start_date = to_date
                if to_date > self.end_date:
                    done = True
        return _values

    def get_crosstab_parsed_columns(self):
        """
        Return a list of the columns analyzed , with reference to computation field and everything
        :return:
        """
        report_columns = self.crosstab_columns or []
        ids = list(self.crosstab_ids)
        if self.crosstab_compute_reminder:
            ids.append('----')
        output_cols = []
        ids_length = len(ids) - 1
        for counter, id in enumerate(ids):
            for col in report_columns:
                # try:
                magic_field_class = field_registry.get_field_by_name(col)
                # except:
                #     magic_field_class = None

                output_cols.append({
                    'name': f'{col}CT{id}',
                    'verbose_name': self.get_crosstab_field_verbose_name(magic_field_class, self.crosstab_model, id),
                    'ref': magic_field_class,
                    'id': id,
                    'model': self.crosstab_model,
                    'is_reminder': counter == ids_length
                })

        return output_cols

    def get_crosstab_field_verbose_name(self, computation_class, model, id):
        """
        Hook to change the crosstab field verbose name, default it delegate this function to the ReportField
        :param computation_class: ReportField Class
        :param model: the model name as string
        :param id: the current crosstab id
        :return: a verbose string
        """
        return computation_class.get_crosstab_field_verbose_name(model, id)



import datetime

import simplejson as json
from django.conf import settings
from django.http import HttpResponse
from django.utils.encoding import force_text
from django.utils.functional import Promise
from django.views.generic import FormView

from .app_settings import SLICK_REPORTING_DEFAULT_END_DATE, SLICK_REPORTING_DEFAULT_START_DATE
from .form_factory import report_form_factory
from .generator import ReportGenerator


class SampleReportView(FormView):
    group_by = None
    columns = None

    time_series_pattern = ''
    time_series_columns = None

    date_field = 'doc_date'

    swap_sign = False

    report_generator_class = ReportGenerator

    report_model = None

    base_model = None
    limit_records = None

    queryset = None

    chart_settings = None
    """
    A list of chart settings objects instructing front end on how to plot the data.
    
    """

    template_name = 'slick_reporting/simple_report.html'

    def get(self, request, *args, **kwargs):
        form_class = self.get_form_class()
        self.form = self.get_form(form_class)
        if self.form.is_valid():
            report_data = self.get_report_results()
            if request.is_ajax():
                return self.ajax_render_to_response(report_data)

            return self.render_to_response(self.get_context_data(report_data=report_data))

        return self.render_to_response(self.get_context_data())

    def ajax_render_to_response(self, report_data):
        return HttpResponse(self.serialize_to_json(report_data),
                            content_type="application/json")

    def serialize_to_json(self, response_data):
        """ Returns the JSON string for the compiled data object. """

        def date_handler(obj):
            if type(obj) is datetime.datetime:
                return obj.strftime('%Y-%m-%d %H:%M')
            elif hasattr(obj, 'isoformat'):
                return obj.isoformat()
            elif isinstance(obj, Promise):
                return force_text(obj)

        indent = None
        if settings.DEBUG:
            indent = 4

        return json.dumps(response_data, indent=indent, use_decimal=True, default=date_handler)

    def get_form_class(self):
        """
        Automatically instantiate a form based on details provided
        :return:
        """
        return self.form_class or report_form_factory(self.report_model)

    def get_form_kwargs(self):
        """
        Returns the keyword arguments for instantiating the form.
        """
        kwargs = {
            'initial': self.get_initial(),
            'prefix': self.get_prefix(),
        }

        if self.request.method in ('POST', 'PUT'):
            kwargs.update({
                'data': self.request.POST,
                'files': self.request.FILES,
            })
        elif self.request.method in ('GET', 'PUT'):

            # elif self.request.GET:
            kwargs.update({
                'data': self.request.GET,
                'files': self.request.FILES,
            })
        return kwargs

    def get_report_generator(self, queryset, for_print):
        q_filters, kw_filters = self.form.get_filters()
        return self.report_generator_class(self.report_model,
                                           kwargs_filters=kw_filters,
                                           date_field=self.date_field,
                                           main_queryset=queryset,
                                           print_flag=for_print,
                                           limit_records=self.limit_records, swap_sign=self.swap_sign,
                                           columns=self.columns,
                                           group_by=self.group_by,
                                           time_series_pattern=self.time_series_pattern,
                                           time_series_columns=self.time_series_columns,
                                           )

    def get_columns_data(self, columns):
        """
        Hook to get the columns information to front end
        :param columns:
        :return:
        """
        # columns = report_generator.get_list_display_columns()
        data = []
        for col in columns:
            data.append({
                'name': col['name'],
                'verbose_name': col['verbose_name'],
                'visible': col.get('visible', True),
                'type': col.get('type', 'text')
            })
        return data

    def get_report_results(self, for_print=False):
        """
        Gets the reports Data, and, its meta data used by datatables.net and highcharts
        :return: JsonResponse
        """

        queryset = self.get_queryset()
        report_generator = self.get_report_generator(queryset, for_print)
        data = report_generator.get_report_data()
        data = self.filter_results(data, for_print)
        data = {
            'report_slug': self.get_report_slug(),
            'data': data,
            'columns': self.get_columns_data(report_generator.get_list_display_columns()),
            'metadata': self.get_metadata(generator=report_generator),
            'chart_settings': self.get_chart_settings()
        }
        return data

    def get_metadata(self, generator):
        """
        A hook to send data about the report for front end which can later be used in charting
        :return:
        """
        time_series_columns = generator.get_time_series_parsed_columns()
        metadata = {
            'time_series_pattern': self.time_series_pattern,
            'time_series_column_names': [x['name'] for x in time_series_columns],
            'time_series_column_verbose_names': [x['verbose_name'] for x in time_series_columns]
        }
        return metadata

    def get_chart_settings(self):
        """setting the chart id.. can be better """
        output = []
        for i, x in enumerate(self.chart_settings or []):
            x['id'] = f"{x['type']}-{i}"
            output.append(x)
        return output


    def get_queryset(self):
        return self.queryset or self.report_model.objects

    def filter_results(self, data, for_print=False):
        """
        Hook to Filter results based on computed data (like eliminate __balance__ = 0, etc)
        :param data: List of objects
        :param for_print: is print request
        :return: filtered data
        """
        return data

    @classmethod
    def get_report_slug(cls):
        return cls.__name__.lower()

    def get_initial(self):
        # todo revise why not actually displaying datetime on screen
        return {
            'start_date': SLICK_REPORTING_DEFAULT_START_DATE,
            'end_date': SLICK_REPORTING_DEFAULT_END_DATE
        }


from slick_reporting.generator import ReportGenerator
from .models import OrderLine


class GenericGenerator(ReportGenerator):
    report_model = OrderLine
    date_field = 'order__date_placed'

    # here is the meat and potatos of the report,
    # we group the sales per client , we display columns slug and title (of the `base_model` defied above
    # and we add the magic field `__balance__` we compute the client balance.
    group_by = 'client'
    columns = ['slug', 'name', '__balance__']


class GeneratorWithAttrAsColumn(GenericGenerator):
    group_by = 'client'

    columns = ['get_data', 'slug', 'name']

    def get_data(self, obj):
        return ''

    get_data.verbose_name = 'get_data_verbose_name'


class CrosstabOnClient(GenericGenerator):
    group_by = 'product'
    columns = ['name', '__total_quantity__']
    crosstab_model = 'client'
    crosstab_columns = ['__total_quantity__']


#
from django.utils.translation import ugettext_lazy as _

from slick_reporting.generator import ReportGenerator
from .models import Client, SimpleSales, Product


class ClientTotalBalance(ReportGenerator):

    report_model = SimpleSales
    date_field = 'doc_date'
    group_by = 'client'
    columns = ['slug', 'name', '__balance__', '__total__']


class ClientTotalBalancesOrdered(ClientTotalBalance):
    report_slug = None
    default_order_by = '__balance__'


class ClientTotalBalancesOrderedDESC(ClientTotalBalance):
    report_slug = None
    default_order_by = '-__balance__'


class ProductTotalSales(ReportGenerator):
    report_model = SimpleSales
    date_field = 'doc_date'
    group_by = 'product'
    columns = ['slug', 'name', '__balance__', '__balance_quan__']


class ClientList(ReportGenerator):
    report_title = _('Our Clients')

    # report_slug = 'client_list'
    base_model = Client
    report_model = SimpleSales

    group_by = 'client'
    columns = ['slug', 'name']


class ProductClientSales(ReportGenerator):
    base_model = Client
    report_model = SimpleSales

    report_slug = 'client_sales_of_products'
    report_title = _('Client net sales for each product')
    must_exist_filter = 'client_id'
    header_report = ClientList

    group_by = 'product'
    columns = ['slug', 'name', '__balance_quan__', '__balance__']


class ProductSalesMonthlySeries(ReportGenerator):
    base_model = Product
    report_model = SimpleSales
    report_title = _('Product Sales Monthly')

    form_settings = {
        'group_by': 'product',
        'group_columns': ['slug', 'name'],

        'time_series_pattern': 'monthly',
        'time_series_columns': ['__balance_quan__', '__balance__'],
    }

    group_by = 'product'
    columns = ['slug', 'name']
    time_series_pattern = 'monthly',
    time_series_columns = ['__balance_quan__', '__balance__']

    chart_settings = [
        {
            'id': 'movement_column',
            'name': _('comparison - column'),
            'settings': {
                'chart_type': 'column',
                'name': _('{product} Avg. purchase price '),
                'sub_title': _('{date_verbose}'),
                'y_sources': ['__balance__'],
                'series_names': [_('Avg. purchase price')],
            }
        },
        {
            'id': 'movement_line',
            'name': _('comparison - line'),
            'settings': {
                'chart_type': 'line',
                'name': _('{product} Avg. purchase price '),
                'sub_title': _('{date_verbose}'),
                'y_sources': ['__balance__'],
                'series_names': [_('Avg. purchase price')],
            }
        },
    ]


class ClientReportMixin:
    base_model = Client
    report_model = SimpleSales


class ClientSalesMonthlySeries(ReportGenerator):
    report_model = SimpleSales
    date_field = 'doc_date'

    group_by = 'client'
    columns = ['slug', 'name']
    time_series_pattern = 'monthly'
    time_series_columns = ['__debit__', '__credit__', '__balance__', '__total__']


#


class ClientDetailedStatement(ReportGenerator):
    report_model = SimpleSales
    date_field = 'doc_date'
    group_by = None
    columns = ['slug', 'doc_date', 'product__name', 'quantity', 'price', 'value']


class ClientDetailedStatement2(ReportGenerator):
    report_title = _('client statement')
    base_model = Client
    report_model = SimpleSales

    header_report = ClientList
    must_exist_filter = 'client_id'

    form_settings = {
        'group_by': '',
        'group_columns': ['slug', 'doc_date', 'doc_type', 'product__title', 'quantity', 'price', 'value'],
    }
    group_by = None
    columns = ['slug', 'doc_date', 'doc_type', 'product__title', 'quantity', 'price', 'value']


class ProductClientSalesMatrix(ReportGenerator):
    report_model = SimpleSales
    date_field = 'doc_date'

    group_by = 'product'
    columns = ['slug', 'name']

    crosstab_model = 'client'
    crosstab_columns = ['__total__']


class GeneratorClassWithAttrsAs(ReportGenerator):
    columns = ['get_icon', 'slug', 'name']


class ClientTotalBalancesWithShowEmptyFalse(ClientTotalBalance):
    report_slug = None
    default_order_by = '-__balance__'
    show_empty_records = False



from django.db.models import Sum
from django.utils.translation import ugettext_lazy as _

from .decorators import report_field_register
from .helpers import get_calculation_annotation
from .registry import field_registry


class BaseReportField(object):
    """
    Computation field responsible for making the calculation unit
    """

    name = ''
    """The name to be used in the ReportGenerator"""

    calculation_field = 'value'
    """the Field to compute on"""

    calculation_method = Sum
    """The computation Method"""

    verbose_name = None
    """Verbose name to be used in front end when needed"""

    requires = None
    """This can be a list of sibling classes,
    they will be asked to compute and their value would be available to you in the `resolve` method
    requires = [BasicCalculationA, BasicCalculationB]
    """

    type = 'number'
    """Just a string describing what this computation field return"""

    report_model = None
    group_by = None
    plus_side_q = None
    minus_side_q = None

    _require_classes = None
    _debit_and_credit = True

    def __init__(self, plus_side_q=None, minus_side_q=None,
                 report_model=None,
                 qs=None,
                 calculation_field=None, calculation_method=None, date_field='', group_by=None):
        super(BaseReportField, self).__init__()
        self.date_field = date_field
        self.report_model = self.report_model or report_model
        self.calculation_field = calculation_field if calculation_field else self.calculation_field
        self.calculation_method = calculation_method if calculation_method else self.calculation_method
        self.plus_side_q = self.plus_side_q or plus_side_q
        self.minus_side_q = self.minus_side_q or minus_side_q
        self.requires = self.requires or []
        self.group_by = self.group_by or group_by
        self._cache = None, None, None
        self._require_classes = [field_registry.get_field_by_name(x) for x in self.requires]

        if not self.plus_side_q and not self.minus_side_q:
            self._debit_and_credit = False

    @classmethod
    def _get_required_classes(cls):
        requires = cls.requires or []
        return [field_registry.get_field_by_name(x) for x in requires]

    def apply_q_plus_filter(self, qs):
        return qs.filter(*self.plus_side_q)

    def apply_q_minus_filter(self, qs):
        return qs.filter(*self.minus_side_q)

    def apply_aggregation(self, queryset, group_by=''):
        annotation = self.calculation_method(self.calculation_field)
        if group_by:
            queryset = queryset.values(group_by).annotate(annotation)
        else:
            queryset = queryset.aggregate(annotation)
        return queryset

    def prepare(self, q_filters=None, kwargs_filters=None, **kwargs):
        """
        This is the first hook where you can customize the calculation away from the Django Query aggregation method
        This method et called with all available parameters , so you can prepare the results for the whole set and save
        it in a local cache (like self._cache) .
        The flow will later call the method `resolve`,  giving you the id, for you to return it respective calculation

        :param q_filters:
        :param kwargs_filters:
        :param kwargs:
        :return:
        """
        kwargs_filters = kwargs_filters or {}

        dep_values = self._prepare_dependencies(q_filters, kwargs_filters.copy())

        queryset = self.get_queryset()
        if q_filters:
            queryset = queryset.filter(*q_filters)
        if kwargs_filters:
            queryset = queryset.filter(**kwargs_filters)

        if self.plus_side_q:
            queryset = self.apply_q_plus_filter(queryset)
        debit_results = self.apply_aggregation(queryset, self.group_by)

        credit_results = None
        if self._debit_and_credit:
            queryset = self.get_queryset()
            if kwargs_filters:
                queryset = queryset.filter(**kwargs_filters)
            if q_filters:
                queryset = queryset.filter(*q_filters)
            if self.minus_side_q:
                queryset = self.apply_q_minus_filter(queryset)

            credit_results = self.apply_aggregation(queryset, self.group_by)

        self._cache = debit_results, credit_results, dep_values
        return debit_results, credit_results, dep_values

    def get_queryset(self):
        queryset = self.report_model.objects
        return queryset

    def get_annotation_name(self):
        """
        Get the annotation per the database
        :return: string used ex:
        """
        return get_calculation_annotation(self.calculation_field, self.calculation_method)

    def _prepare_dependencies(self, q_filters=None, extra_filters=None, ):
        values = {}
        for dep_class in self._require_classes:
            dep = dep_class(self.plus_side_q, self.minus_side_q, self.report_model,
                            date_field=self.date_field, group_by=self.group_by)
            values[dep.name] = {'results': dep.prepare(q_filters, extra_filters),
                                'instance': dep}
        return values

    def resolve(self, current_obj):
        '''
        Reponsible for getting the exact data from the prepared value
        :param cached: the returned data from prepare
        :param current_obj:
        :return: a solid number or value
        '''
        cached = self._cache
        debit_value, credit_value = self.extract_data(cached, current_obj)
        dependencies_value = self._resolve_dependencies(current_obj)

        return self.final_calculation(debit_value, credit_value, dependencies_value)

    def get_dependency_value(self, current_obj, name=None):
        """
        Get the values of the ReportFields specified in `requires`

        :param current_obj: the current object which we want the calculation for
        :param name: Optional, the name of the specific dependency you want.

        :return: a dict containing dependencies names as keys and their calculation as values
                 or a specific value if name is specified.
        """
        values = self._resolve_dependencies(current_obj)
        if name:
            return values.get(name)
        return values

    def _resolve_dependencies(self, current_obj):

        dep_results = {}
        cached_debit, cached_credit, dependencies_value = self._cache
        dependencies_value = dependencies_value or {}
        for d in dependencies_value.keys():
            d_instance = dependencies_value[d]['instance']
            dep_results[d] = d_instance.resolve(current_obj)
        return dep_results

    def extract_data(self, cached, current_obj):
        group_by = self.group_by
        debit_value = 0
        credit_value = 0
        annotation = self.get_annotation_name()

        cached_debit, cached_credit, dependencies_value = cached

        if cached_debit or cached_credit:
            debit = None
            if cached_debit is not None:
                if not group_by:
                    x = cached_debit.keys()[0]
                    debit_value = cached_debit[x]
                else:
                    for i, x in enumerate(cached_debit):
                        if str(x[group_by]) == current_obj:
                            debit = cached_debit[i]
                            break
                    if debit:
                        debit_value = debit[annotation]

            if cached_credit is not None:
                credit = None
                if cached_credit is not None:
                    if not group_by:
                        x = cached_credit.keys()[0]
                        credit_value = cached_credit[x]
                    else:
                        for i, x in enumerate(cached_credit):
                            if str(x[group_by]) == current_obj:
                                credit = cached_credit[i]
                                break
                        if credit:
                            credit_value = credit[annotation]
        return debit_value, credit_value

    def final_calculation(self, debit, credit, dep_dict):
        debit = debit or 0
        credit = credit or 0
        return debit - credit

    @classmethod
    def get_full_dependency_list(cls):
        """
        Get the full Hirearchy of dependencies and dependencies dependency.
        :return: List of dependecies classes
        """

        def get_dependency(field_class):

            dependencies = field_class._get_required_classes()
            klasses = []
            for klass in dependencies:
                klasses.append(klass)
                other = get_dependency(klass)
                if other:
                    klasses += other
            return klasses

        return get_dependency(cls)

    @classmethod
    def get_crosstab_field_verbose_name(cls, model, id):
        """
        Construct a verbose name for the crosstab field
        :param model: the model name
        :param id: the id of the current crosstab object
        :return: a verbose string
        """
        return f'{cls.verbose_name} {model} {id}'

    @classmethod
    def get_time_series_field_verbose_name(cls, date_period):
        """
        Sent the column data to construct a verbose name.
        Default implemenetation is column name + the end date %Y%m%d
        :param column_name: the computation field_name
        :param date_period: a tuple of (start_date, end_date)
        :return: a verbose string
        """
        return f'{cls.verbose_name} {date_period[1].strftime("%Y%m%d")}'


class FirstBalanceField(BaseReportField):
    name = '__fb__'
    verbose_name = _('first balance')

    def prepare(self, q_filters=None, extra_filters=None, **kwargs):
        extra_filters = extra_filters or {}

        from_date_value = extra_filters.get(f'{self.date_field}__gt')
        extra_filters.pop(f'{self.date_field}__gt', None)
        extra_filters[f'{self.date_field}__lte'] = from_date_value
        return super(FirstBalanceField, self).prepare(q_filters, extra_filters)


field_registry.register(FirstBalanceField)


class TotalReportField(BaseReportField):
    name = '__total__'
    verbose_name = _('total')
    requires = ['__debit__', '__credit__']


field_registry.register(TotalReportField)


class BalanceReportField(BaseReportField):
    name = '__balance__'
    verbose_name = _('balance')
    requires = ['__fb__']

    def final_calculation(self, debit, credit, dep_dict):
        fb = dep_dict.get('__fb__')
        debit = debit or 0
        credit = credit or 0
        fb = fb or 0
        return fb + debit - credit


field_registry.register(BalanceReportField)


class CreditReportField(BaseReportField):
    name = '__credit__'
    verbose_name = _('credit')

    def final_calculation(self, debit, credit, dep_dict):
        return credit


field_registry.register(CreditReportField)


class DebitReportField(BaseReportField):
    name = '__debit__'
    verbose_name = _('debit')

    def final_calculation(self, debit, credit, dep_dict):
        return debit


field_registry.register(DebitReportField)


@report_field_register
class DocCount(BaseReportField):
    name = '__doc_count__'
    verbose_name = _('document count')


@report_field_register
class LineCount(BaseReportField):
    name = '__line_count__'
    verbose_name = 'Line Count'


# @report_field_register
# class DocValue

class TotalQTYReportField(BaseReportField):
    name = '__total_quantity__'
    verbose_name = _('total QTY')
    calculation_field = 'quantity'


field_registry.register(TotalQTYReportField)


class FirstBalanceQTYReportField(FirstBalanceField):
    name = '__fb_quan__'
    verbose_name = _('first balance QTY')
    calculation_field = 'quantity'


field_registry.register(FirstBalanceQTYReportField)


class BalanceQTYReportField(BaseReportField):
    name = '__balance_quan__'
    verbose_name = _('balance QTY')
    calculation_field = 'quantity'
    requires = ['__fb_quan__']

    def final_calculation(self, debit, credit, dep_dict):
        # Use `get` so it fails loud if its not there
        fb = dep_dict.get('__fb_quan__')
        fb = fb or 0
        return fb + debit - credit


field_registry.register(BalanceQTYReportField)



from slick_reporting.views import SampleReportView
from .models import OrderLine


class MonthlyProductSales(SampleReportView):
    report_model = OrderLine
    date_field = 'date_placed'  # or 'order__date_placed'
    group_by = 'product'
    columns = ['name', 'sku']
    time_series_pattern = 'monthly'
    time_series_columns = ['__total_quantity__']



from django.db import models
from django.urls import reverse_lazy

from django.utils.translation import ugettext_lazy as _


class Product(models.Model):
    slug = models.CharField(max_length=200, verbose_name=_('Slug'))
    name = models.CharField(max_length=200, verbose_name=_('Name'))
    sku = models.CharField(max_length=200, default='', blank=True)
    notes = models.TextField()

    class Meta:
        verbose_name = _('Product')
        verbose_name_plural = _('Products')


class Client(models.Model):
    slug = models.CharField(max_length=200, verbose_name=_('Slug'))

    name = models.CharField(max_length=200, verbose_name=_('Name'))
    email = models.EmailField(blank=True)
    notes = models.TextField()

    class Meta:
        verbose_name = _('Client')
        verbose_name_plural = _('Clients')


class SimpleSales(models.Model):
    slug = models.SlugField()
    doc_date = models.DateTimeField(_('date'), db_index=True)
    client = models.ForeignKey(Client, on_delete=models.CASCADE)
    product = models.ForeignKey(Product, on_delete=models.CASCADE)
    quantity = models.DecimalField(_('quantity'), max_digits=19, decimal_places=2, default=0)
    price = models.DecimalField(_('price'), max_digits=19, decimal_places=2, default=0)
    value = models.DecimalField(_('value'), max_digits=19, decimal_places=2, default=0)

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.value  = self.quantity * self.price
        super().save(force_insert, force_update, using, update_fields)

    class Meta:
        verbose_name = _('Sale')
        verbose_name_plural = _('Sales')

#
# class Invoice(BaseMovementInfo):
#     client = models.ForeignKey(Client, on_delete=models.CASCADE)
#
#     @classmethod
#     def get_doc_type(cls):
#         return 'sales'
#
#
# class InvoiceLine(QuanValueMovementItem):
#     invoice = models.ForeignKey(Invoice, on_delete=models.CASCADE)
#
#     product = models.ForeignKey(Product, on_delete=models.CASCADE)
#     client = models.ForeignKey(Client, on_delete=models.CASCADE)
#
#     @classmethod
#     def get_doc_type(cls):
#         return 'sales'
#
#
# class Journal(BaseMovementInfo):
#     data = models.CharField(max_length=100, null=True, blank=True)
#
#     @classmethod
#     def get_doc_type(cls):
#         return 'journal-sales'
#
#
# class JournalItem(BaseMovementItemInfo):
#     journal = models.ForeignKey(Journal, on_delete=models.CASCADE)
#     client = models.ForeignKey(Client, on_delete=models.CASCADE)
#     data = models.CharField(max_length=100, null=True, blank=True)
#
#     @classmethod
#     def get_doc_type(cls):
#         return 'journal-sales'
#
#
# class JournalWithCriteria(Journal):
#     class Meta:
#         proxy = True
#

# Vanilla models

class Order(models.Model):
    date_placed = models.DateTimeField(auto_created=True)
    client = models.ForeignKey(Client, null=True, on_delete=models.CASCADE)


class OrderLine(models.Model):
    date_placed = models.DateTimeField(auto_created=True)
    product = models.ForeignKey(Product, on_delete=models.CASCADE)
    order = models.ForeignKey(Order, on_delete=models.CASCADE)
    quantity = models.PositiveIntegerField(default=0)
    client = models.ForeignKey(Client, null=True, on_delete=models.CASCADE)
